#include <fstream>
#include "logMap.h"

// ****************************************************************************
// procedure createFirstGrid
//   this procedure divides interval p onto n (almost) equal parts and sets 
//   status of each as unknown (i.e. we know nothing about existence of stable orbit)
//
// input: n - number of parts on which we divide interval
//        p - interval to divide 
//
// output: c - container which has been filled by subintervals with associated 
//             information about stability
// ****************************************************************************
void createFirstGrid(DataContainer& c, int n, Interval p)
{
  RoundClass::roundNearest();
  Real leftEnd = p.leftBound();
  Real diam = p.rightBound()-p.leftBound();
  
  for(int i=1;i<n;++i)
  {
  // this is only approximate value, it is not necessary to have it exact (as interval)
    Real rightEnd = leftEnd + diam/n;
    c.insert( ParameterInfo<Interval>(Interval(leftEnd,rightEnd)) );
    // here we guarantee that whole interval p is covered 
    leftEnd = rightEnd;
  }
  // last interval will have end p.rightBound()
  c.insert( ParameterInfo<Interval>(Interval(leftEnd,p.rightBound())) );
}

// ****************************************************************************
// procedure saveDataContainer
//
// input: filename - a file to which data will be saved
//        c - container which contains data so save
//        clearFile - if true, the file will be overwrtitten
//                    otherwise, data will be saved at the end  
//
// ****************************************************************************
void saveDataContainer(const char filename[], DataContainer& c, bool clearFile)
{
  RoundClass::roundNearest();
  std::ofstream aFile;
  if(clearFile)
    aFile.open(filename);
  else
  {
    aFile.open(filename,std::ios::app);
    aFile << std::endl;
  }  
  aFile.precision(21);

  DataContainer::iterator b = c.begin(), e=c.end();
  std::cout << "writing data to a file " << filename << " ... " << std::flush;

  int check = false;
  Interval r;
  while(b!=e)
  {
    if(check)
    {
      if(r.rightBound()!=b->c.leftBound())
      {
        std::cout << "Overlaped subintervals!!!! ----------------------- \n";
        std::cout << r  << "\n" << b->c << std::endl;
        exit(0);
      }
    }
    r = b->c;
    check = true;

    aFile << (*b);
    ++b;
    if(b!=e)
      aFile << std::endl;
  }
  aFile.close();
  std::cout << "done" << std::endl;
}

// ****************************************************************************
// procedure readDataFromFile
//
// input: - filename - file which contains data in proper format
//        - fromLine - number of line in the file from which we start to read data (counted from zero).
//          We can skip some number of lines. Usefull, when whole file cannot be 
//          read at once because of the memory usage.
//        - maxNumberOfLines - how many lines to read. When equal to zero 
//          all data will be readed
//
// ouput: c - containers with inserted data from the file.
//        Also an information about the number of lines which has been readed is returned. 
// ****************************************************************************
unsigned long readDataFromFile(
  const char filename[], 
  DataContainer& c, 
  int fromLine, 
  unsigned long maxNumberOfLines
  )
{
  RoundClass::roundNearest();
  std::ifstream aFile;

  aFile.open(filename);

  std::string s;
  while(fromLine>0 && !aFile.eof())
  {
    getline(aFile,s);
    fromLine--;
  } 
  
  // maxNumberOfLines cannot be negative
  // therefore if it is equal to zer0, then ot could be seen  
  // as an infinity for operator++
  unsigned long lines=0;
  while(!aFile.eof()){
    ParameterInfo<Interval> p;
    aFile >> p;
    c.insert(p);

    ++lines;
    if(maxNumberOfLines==lines)
      break;
  };
  aFile.close();
  return lines;
}


// ****************************************************************************
// procedure computeTotalWidth
//
// input: filename - name of file with data
//
// ouput: the procedure reads data form the file and computes range on which 
//        the stability has been verified. Result is printed to the standard output.
// ****************************************************************************
void computeTotalWidth(const char filename[], bool separatePeriods)
{
  typedef std::pair<unsigned int,Interval> PerInfo;
  typedef std::map<unsigned int, PerInfo> PerSet;

  PerSet perSet;
  unsigned int previous=0;
  RoundClass::roundNearest();
  std::ifstream aFile;
  std::set<int> s;
  std::set< std::pair<Interval,int> > pd;

  aFile.open(filename);
  ParameterInfo<Interval> p;
  Interval size(0);
  std::cout << "computing stability range ... " << std::endl;
  bool check=false;
  Interval r=0;
  ParameterFlag f=unknown;

  while(!aFile.eof())
  {
    aFile >> p;
    if(p.flag==stability || p.flag==periodDoubling)
    {
      size += p.c.right() - p.c.left();
      s.insert(p.period);
      if(separatePeriods && p.flag==stability)
      {
        if(perSet.find(p.period)!=perSet.end())
          perSet[p.period].second += p.c.right() - p.c.left();
        else
          perSet[p.period] = PerInfo(0,p.c.right() - p.c.left());
        if(previous!=p.period)
          (perSet[p.period].first)++;
        previous = p.period;
      }
    }
    
    if(check)
    {
      if(f==periodDoubling && p.flag==periodDoubling)
      {
        std::cout << "Two period doublings connected!!!-----------------\n";
        std::cout << p << std::endl;
        exit(0);
      }
      if(p.c.leftBound()!=r.rightBound() && p.c.leftBound()!=0)
      {
        std::cout << "Overlaped subintervals!!!! ----------------------- \n";
        std::cout << r  << "\n" << p.c << std::endl;
        exit(0);
      }
    }else
    {
      check=true;
    }
    r = p.c;
    if(p.flag==periodDoubling)
      pd.insert(std::pair<Interval,int>(p.c,p.period));
  }
  aFile.close();

  std::cout << "stability range: " << size << "\nverified periods: { ";
  std::set<int>::iterator i=s.begin(), j=s.end();
  while(i!=j)
  {
    std::cout << *i << " ";
    ++i;
  }
  std::cout << "}\n";

  std::cout << "verified period doubling (" << pd.size() << "): {\n";
  std::set< std::pair<Interval,int> >::iterator b=pd.begin(), e=pd.end();
  while(b!=e)
  {
    std::cout << "  " << b->second << "->" << 2*(b->second) << ": " << b->first << std::endl;
    ++b;
  }
  std::cout << "}" << std::endl << std::flush;
  if(separatePeriods)
  {
    std::cout << "verified periods (" << perSet.size() << "): {\n";
    PerSet::iterator g=perSet.begin(), h=perSet.end();
    while(g!=h)
    {
      std::cout << g->first << " & " << g->second.first << " & " << g->second.second << "\\\\\n";
      ++g;
    }
  }
}

// ****************************************************************************
// procedure releaseMemoty
// 
// input: c - data container
// 
// output: c - contains allocated interval vectors for each node, for which stability
//             or period doubling has been verified. This procedure releases this memory.
// ****************************************************************************
void releaseMemory(DataContainer& c)
{
  DataContainer::iterator b = c.begin(), e=c.end();
  while(b!=e)
  {
    if(b->flag==stability || b->flag==periodDoubling)
    {
      delete b->enclosure;
    }
    ++b;
  }
}

void collapseData(DataContainer& c)
{
  DataContainer::iterator b = c.begin(), e=c.end();

  while(b!=e)
  {
    if(b->flag==stability && b->enclosure->dimension()>8)
    {
      DataContainer::iterator it = b;
      ParameterInfo<Interval> before = *b;
      int period = b->enclosure->dimension();
      IVector enc(b->enclosure->dimension());
      ++b;
      while( b!=e && b->flag==stability)
      {
        capd::vectalg::intervalHull(*(b->enclosure),*(before.enclosure),enc);
        if(!checkDisjointSets(enc.begin(),period))
          break;
        *(before.enclosure) = enc;
        delete b->enclosure;
        before.c.setRightBound(b->c.rightBound());
        ++b;
      }
      c.erase(it,b);
      c.insert(before);
    }else
      ++b;
  }
}


