#ifndef _LOGISTIC_MAP_H_
#define _LOGISTIC_MAP_H_

// this class represenst a logistic map c*x*(1-x)
template<class IntervalT>
class LogisticMap{
public:
  typedef IntervalT IntervalType;
  typedef typename IntervalT::BoundType FloatType;
  
  void setParameter(IntervalType _c)
  {
    c = _c;
    midC = c.mid().leftBound();
  }
  
  const IntervalType& getParameter() const
  {
    return c;
  }

  // nonrigorous value of function
  FloatType operator()(FloatType x)
  {
    return midC*x*(1.-x);
  }
  
  // nonrigorous derivative of function
  FloatType operator[](FloatType x)
  {
    return midC*(1.-2.*x);
  }

  // rigorous value of function
  IntervalType operator()(const IntervalType& x)
  {
    return c*(0.25 - capd::intervals::power(x-0.5,2));
  }
  
  // rigorous derivative of function
  IntervalType operator[](const IntervalType& x)
  {
    return 2*c*(0.5-x);
  }

  // rigorous routine which computes partial derivatives for iteration of Logistic map
  void composeDerivatives(Derivatives& d)
  {
    d.d3xxx = d.c*(-6.*d.dx*d.d2xx + (1. - 2.*d.x)*d.d3xxx);
    d.d2xx = d.c*(-2*capd::intervals::power(d.dx,2) + (1. - 2.*d.x)*d.d2xx);
    d.d2cx = (1. - 2.*d.x - 2.*d.c*d.dc)*d.dx + d.c*(1. - 2.*d.x)*d.d2cx;
    d.dx = d.c*(1. - 2.*d.x)*d.dx;
    d.dc = (1. - d.x)*d.x + d.c*(1. - 2.*d.x)*d.dc;
    d.x = d.c*(0.25 - capd::intervals::power(d.x-0.5,2.));
  }

  void composeDerivatives(RealDerivatives& d)
  {
    d.d3xxx = d.c*(-6.*d.dx*d.d2xx + (1. - 2.*d.x)*d.d3xxx);
    d.d2xx = d.c*(-2*d.dx*d.dx + (1. - 2.*d.x)*d.d2xx);
    d.d2cx = (1. - 2.*d.x - 2.*d.c*d.dc)*d.dx + d.c*(1. - 2.*d.x)*d.d2cx;
    d.dx = d.c*(1. - 2.*d.x)*d.dx;
    d.dc = (1. - d.x)*d.x + d.c*(1. - 2.*d.x)*d.dc;
    d.x = d.c*(0.25 - (d.x-0.5)*(d.x-0.5));
  }

  static void composeDerivatives(const Interval& x, const Interval& c, Interval& dx, Interval& dc)
  {
    dx = c*(1. - 2.*x)*dx;
    dc = (1. - x)*x + c*(1. - 2.*x)*dc;
  }

  // necessary for general algorithm (we wish to use it for the other logistic like maps)
  const static FloatType criticalPoint()
  {
    return 0.5;
  }

  IntervalType c;
  FloatType midC;
};

#endif
