#include "logMap.h"
#include "ParameterInfo.h"
#include "backwardShooting.h"

//*******************************************************************************************
// procedure checkDisjointSets
//
// input: rigorous bound for periodic trajectory
//        period of this orbit
//
// output: true if period is minimal period, i.e. periodicTrajectory[0] is disjoint with 
//            with other coefficients
//         false otherwise
//*******************************************************************************************
bool checkDisjointSets(const Interval periodicTrajectory[], unsigned long period)
{
  const Interval& p = *periodicTrajectory;
  for(int i=1; i<period;++i)
  {
    if( !(period%i) &&
        p.leftBound()<=periodicTrajectory[i].rightBound() && 
        p.rightBound()>=periodicTrajectory[i].leftBound()
      )
    {
      #ifdef _DEBUG_LM_
        std::cout << "trajectory might be of lower priod\n";
      #endif
      return false;
    }
  }
  return true;
}

//*******************************************************************************************
// algorithm computeNewtonOperator
// input: f - a logistic map
//        period - expected period of the point
//        center - approximate periodic trajectory
//        X - a set in which we expect the existence of an unique periodic trajectory
//
// output: either rigorous enclosure for stable periodic orbit (in variable o_periodicTrajectory) 
//            and true
//         or false
//
// assumptions - class MapType must provide:
//  - public type IntervalType
//  - public type FloatType
//  - operator() computing value of the function both rigorous and approximate value
//  - operator[] computing derivative of the function both rigorous and approximate value
//*******************************************************************************************
bool computeNewtonOperator(
    LogMap& f, 
    unsigned long period,
    const Interval* center,
    const Interval* X,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& product
  )
{
  // we will compute the Interval Newton Operator for the map F(x1,...,xn) = (x2-f(x1),....,x1-f(xn))  - see Galias paper
  static Interval da[numberOfProcessors*maximumPeriod],
                   dg[numberOfProcessors*maximumPeriod];

  Interval *a = da + threadNumber*maximumPeriod,
           *g = dg + threadNumber*maximumPeriod;

  unsigned long i;  

  // having the center and the set we can compute F and F'
  for(i=0;i<period;++i)
  {
    g[i] = center[(i+1)%period] - f(center[i]);
    a[i] = f[X[i]];
  }
  BackwardShootingResult info = backwardShooting(period,a,center,g,o_periodicTrajectory,product);

  if( info==unstable )
    return false; // impossible to prove stability
  if( info==zeroAtDiagonal )
  {
  // if we cannot use backward shooting,
  // try to apply the interval Krawczyk operator 
    computeKrawczykOperator(f,center,X,a,g,o_periodicTrajectory,period,threadNumber);   
  }

  RowVector refPT(o_periodicTrajectory,period);
  RowVector refX(X,period);
  RowVector refCenter(center,period);

  // after this loop o_periodicTrajectory will contain the Interval Newton or Krawczyk operator
  refPT = refCenter - refPT;

  // we verify if the Newton operator is a subset of X
  if( ! capd::vectalg::subsetInterior(refPT,refX) )
    return false;

  product=1.;
  for(i=0;i<period;++i)
    product *= f[o_periodicTrajectory[i]];
  return true;
}


//*******************************************************************************************
// algorithm verifyStablePeriodicOrbit
// input: f - a logistic map
//        period - expected period of the point
//        center - a mid vector of X
//        X - a cset on which we compute the Interval Newton Operator
//
// output: either rigorous enclosure for stable periodic orbit (in variable o_periodicTrajectory) 
//            and ParameterFlag: stability 
//         or ParameterFlag: convergenceWithoutProof, existenceWithoutStability
//
// assumptions - class MapType must provide:
//  - public type IntervalType
//  - public type FloatType
//  - operator() computing value of the function both rigorous and approximate value
//  - operator[] computing derivative of the function both rigorous and approximate value
//*******************************************************************************************

inline Interval midInterval(const Interval& u)
{
  return u.mid();
}

ParameterFlag verifyStablePeriodicOrbit(
    LogMap& f, 
    unsigned long period,
    const Interval* center,
    const Interval* X,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& o_product
  )
{
  RowVector refPT(o_periodicTrajectory,period);
  RowVector refX(X,period);
  RowVector refCenter(center,period);
  int i=5;
  while(i)
  {
    bool verified = computeNewtonOperator(f,period,center,X,threadNumber,o_periodicTrajectory,o_product);
    if(verified)
    {
      if(checkDisjointSets(o_periodicTrajectory,period))
        break;
      std::copy(refPT.begin(),refPT.end(),refX.begin());
    } else
    {
      if( !capd::vectalg::intersection(refPT,refX,refX) )
        return convergenceWithoutProof;
    }

    if(! capd::vectalg::subset(refCenter,refX) )
    {
      capd::vectalg::mid(refX,refCenter);
    }
    --i;
  }

  // if periodic orbit exists, we must verify if the period is equal to 'period', 
  // i.e. if enclosures for the points on trajectory are pairwise disjoint
     
  if(!i)
    return convergenceWithoutProof;

  if(!(capd::abs(o_product)<1.) )
    return existenceWithoutStability;
  
  return stability;
}


//*******************************************************************************************
// algorithm verifyStablePeriodicOrbit
// input: f - a logistic map
//        period - expected period of the point
//        diam - diameter of the set around periodic point
//        periodicPoint - a candidate for periodic trajectory
//
// output: either rigorous enclosure for stable periodic orbit (in variable o_periodicTrajectory) 
//            and ParameterFlag: stability 
//         or ParameterFlag: convergenceWithoutProof, existenceWithoutStability
//
// assumptions - class MapType must provide:
//  - public type IntervalType
//  - public type FloatType
//  - operator() computing value of the function both rigorous and approximate value
//  - operator[] computing derivative of the function both rigorous and approximate value
//*******************************************************************************************
ParameterFlag verifyStablePeriodicOrbit(
    LogMap& f, 
    unsigned long period,
    Real diameter,
    Real* periodicPoint,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& o_product,
    bool useBrouwer
  )
{
  // we will compute the Interval Newton Operator for the map F(x1,...,xn) = (x2-f(x1),....,x1-f(xn))  - see Galias paper
  static Interval dcenter[numberOfProcessors*maximumPeriod],
                   dX[numberOfProcessors*maximumPeriod];

  unsigned long i;
  Interval diam(-diameter,diameter);
  if(useBrouwer)
  {
    o_periodicTrajectory[0]=periodicPoint[0] + diam;
    for(i=1;i<period;++i)
      o_periodicTrajectory[i] = f(o_periodicTrajectory[i-1]);
    
    if( f(o_periodicTrajectory[period-1]).subset(o_periodicTrajectory[0]) )
    {
      // iterate 10 times the map
      for(i=0;i<=10*period;++i)
        capd::intervals::intersection(o_periodicTrajectory[i%period],f(o_periodicTrajectory[(period+i-1)%period]),o_periodicTrajectory[i%period]);

      o_product=1.;
      for(Interval* j=o_periodicTrajectory;j<o_periodicTrajectory+period;++j)
        o_product *= f[*j];

      if( capd::abs(o_product)<1.  && checkDisjointSets(o_periodicTrajectory,period) )
      {
//        std::cout << "zadzialal brouwer, period=" << period << ", product=" << o_product << std::endl;
        return stability;
      }
    }
  }

  Interval *center = dcenter + threadNumber*maximumPeriod,
           *X = dX + threadNumber*maximumPeriod;

  // we store in center approximate nonrigorous trajectory
  // in X a set on which we verify the newton operator
  // Fcenter is the value of  at center  
  // this will be a center of set for the Interval Newton Method
  for(i=0;i<period;++i)
  {
    center[i] = Interval(periodicPoint[i]);
    X[i] = center[i] + diam;
   }
  
  return verifyStablePeriodicOrbit(f,period,center,X,threadNumber,o_periodicTrajectory,o_product);
}

