#ifndef __BACKWARD_SHOOTING_H__
#define __BACKWARD_SHOOTING_H__

enum BackwardShootingResult {zeroAtDiagonal, succeed, unstable};

//*******************************************************************************************
// algorithm backwardShooting - see Galias' paper
//
// input: a - derivatives of function along trajectory
//        center - candidate for periodic trajectory
//        g - values of funtion f(x_i) - x_{i+1}
//        period - period of possible period orbit
//
// output: succeed if computations succeed, 
//         zeroAtDiagonal if 0 appears at the diagonal
//         unstable if cannot verify stability
//         estimation for Df^{-1}(X)*f(center) in argument o_periodicTrajectory
//         estimation for product of derivatives along the trajectory in the argument product
//*******************************************************************************************
template<class Scalar>
BackwardShootingResult backwardShooting(unsigned long period, const Scalar a[], const Scalar center[], const Scalar g[], Scalar o_periodicTrajectory[], Scalar& product)
{
  product = 1.;
  o_periodicTrajectory[0] = Scalar(0.);
  unsigned long i;
  for(i=0;i<period;++i)
  {
    product *= a[i];
    if(isSingular(product))
    {
      return zeroAtDiagonal;
    }
    if( capd::abs(product)>1e+300)
    {
      return unstable;
    }
    o_periodicTrajectory[0] -= g[i]/product;
  }
  // at the end h1 must be divided by 1-1/product
  Scalar inverseProduct = 1.-1./product;
  if(isSingular(inverseProduct))
  {
    return unstable;
  }
  o_periodicTrajectory[0] /= inverseProduct;
    
  // backward computations of h_k
  for(i=period-1;i>0;i--)
    o_periodicTrajectory[i] = (o_periodicTrajectory[(i+1)%period]-g[i])/a[i];
  return succeed;
}

#endif

