#include <iostream>
#include <set>
#include <cmath>
#include "logMap.h"

using namespace std;
typedef std::set<Real> SuperstableSet;

Real computeValueDerivative(Real c, Real x, Real& dc, int n)
{
  dc = 0.;
  for(int i=0;i<n;++i)
  {
    dc = (1. - x)*x + c*(1. - 2.*x)*dc;
    x = c*(1. - x)*x;
  }
  return x;
}

void removeLowPeriods(SuperstableSet& s, int period)
{
  SuperstableSet::iterator b=s.begin(), e=s.end();
  while(b!=e)
  {
    Real c = *b;
    Real x = 0.5;
    SuperstableSet::iterator temp = b;
    ++b;
    for(int i=0;i<period-1;++i)
    {
      x = c*x*(1-x);
      if(fabs(x-0.5)<1e-14)
      {
        s.erase(temp);
        break;
      }
    }
  }
  
  b=s.begin(), e=s.end();
  while(b!=e)
  {
    Real temp = *b;
    ++b;
    if(b!=e && fabs((*b)-temp)<1e-15)
    {
      s.erase(temp);
    }
  }  
}

void printSet(SuperstableSet& s)
{
  SuperstableSet::iterator b = s.begin(), e = s.end();
  std::cout << "number of detected windows=" << s.size() << std::endl;
  while(b!=e)
  {
    std::cout << "c=" << (*b) << std::endl;
    ++b;
  }
}

void printSet(DataContainer& s)
{
  DataContainer::iterator b = s.begin(), e = s.end();
  while(b!=e)
  {
    std::cout << "c=" << (b->c) << std::endl;
    ++b;
  }
}

void proveSuperstable(LogMap &f, DataContainer& container, SuperstableSet& s, int period, Interval intv)
{
  static Real dtab[numberOfProcessors*maximumPeriod];
  static Interval dintvtab[numberOfProcessors*maximumPeriod];
  int threadNumber = omp_get_thread_num();
//  std::cout << "numer watku = " << threadNumber << std::endl;
  Real *tab = dtab + threadNumber*maximumPeriod;
  Interval *intvtab = dintvtab + threadNumber*maximumPeriod;
  
  SuperstableSet::iterator b = s.begin(), e = s.end();
  Interval product;
  int counter=0;
  unsigned int elements = s.size();

  while(b!=e)
  {
    Real c = *b;
    SuperstableSet::iterator temp = b;
    tab[0] = 0.5;
    Interval param = c+1e-17*Interval(-1,1);
    intersection(param,intv,param);
    f.setParameter(param);
    for(int i=1;i<period;++i)
    {
      tab[i] = f(tab[i-1]);
    }
    Real diameter = 1e-9;    
    ParameterFlag flag = verifyStablePeriodicOrbit(f,period,diameter,tab,threadNumber,intvtab,product,true);
    ++b;
    if(flag!=stability)
    {  
      counter++;      
      s.erase(temp);
    }else
    {
      ParameterInfo<Interval> p(f.getParameter(),stability,intvtab,period);
      container.insert(p);
    }
  }
//  std::cout << "nie sprawdzono: " << counter << " z " << elements << std::endl;
}

void findSuperstable(SuperstableSet& superstable, long long period, long long grid, Interval intv)
{
  Real x = 0.5;
  Real dc;
  Real step = diam(intv).leftBound()/grid;
  int newtonIterations=10;
//  std::cout << "grid=" << grid << std::endl;
  for(long long i=1;i<grid;++i)
  {
    Real c = intv.leftBound() + i*step;
    for(int j=0;j<newtonIterations;++j)
    {
      Real Qx = computeValueDerivative(c,x,dc,period) - x;
      c = c - Qx/dc;
      if(c>= intv.rightBound() || c<=intv.leftBound()) break;
    }
    if(c>= intv.rightBound() || c<=intv.leftBound()) continue;
    if(fabs(computeValueDerivative(c,x,dc,period) - x)<1e-13) // accuracy of periodic orbit
      superstable.insert(c);
  }
}

void computeWindowsWidth(LogMap &f, DataContainer& container, int period, Interval intv)
{ 
  int threadNumber = omp_get_thread_num();
  DataContainer::iterator b = container.begin(), e = container.end();

  static Real dtab[numberOfProcessors*maximumPeriod];
  static Interval dintvtab[numberOfProcessors*maximumPeriod];
  Real *periodicPoint = dtab + threadNumber*maximumPeriod;
  Interval *intvtab = dintvtab + threadNumber*maximumPeriod;

  Interval intvProduct;
  DataContainer newContainer;
  
  int maxGrid=100;
  while(b!=e)
  {
    f.setParameter(b->c);
    unsigned int period = b->period;
    RealDerivatives der((*(b->enclosure))[0].mid().leftBound(),b->c.mid().leftBound());
    for(unsigned int i=0;i<period;++i)
      f.composeDerivatives(der);

    Real deltaC = (der.d2cx + der.d2xx*der.dc/(1.-der.dx));

    int j=maxGrid;
    Real center = (b->c.leftBound()+b->c.rightBound())/2.;
    while(j)
    {
      Real c = center + j*.8/maxGrid/deltaC;
      if(!intv.containsInInterior(c))
      {
        --j;
        continue;
      }
      f.setParameter(c);
      Real product = computeNewtonOperator(f,0.5,period,periodicPoint,10,threadNumber);    
      if(fabs(product)>=1)
      {
        --j;
        continue;
      }
      Interval intvC = c + Interval(-1e-17,1e-17);
      intersection(intvC,intv,intvC);
      f.setParameter(intvC);
      ParameterFlag flag = verifyStablePeriodicOrbit(f,period,1e-9,periodicPoint,threadNumber,intvtab,intvProduct,true);
      if(flag==stability)
      {
        ParameterInfo<Interval> p(f.getParameter(),stability,intvtab,period);
        newContainer.insert(p);
        j=1;
      }
      --j;
    }
    
    j=maxGrid;
    while(j)
    {
      Real c = center - j*1.5/maxGrid/deltaC;
      if(!intv.containsInInterior(c))
      {
        --j;
        continue;
      }
      f.setParameter(c);
      Real product = computeNewtonOperator(f,0.5,period,periodicPoint,10,threadNumber);    
      if(fabs(product)>=1)
      {
        --j;
        continue;
      }
      Interval intvC = c + Interval(-1e-17,1e-17);
      intersection(intvC,intv,intvC);
      f.setParameter(intvC);
      ParameterFlag flag = verifyStablePeriodicOrbit(f,period,1e-9,periodicPoint,threadNumber,intvtab,intvProduct,true);
      if(flag==stability)
      {
        ParameterInfo<Interval> p(f.getParameter(),stability,intvtab,period);
        newContainer.insert(p);
        j=1;
      }
      --j;
    }

    ++b;
  }
  container.insert(newContainer.begin(),newContainer.end());
}

void insertUnknownInfo(DataContainer& c, Interval intv)
{
  DataContainer::iterator b=c.begin(), e=c.end();
  Real left = intv.leftBound();

  while(b!=e)
  {
    Real right = b->c.leftBound();
    if(left>=right)
    {
      if(left>b->c.rightBound())
      {
        throw std::runtime_error("Subdivison error");
      }
    }
    if(left<right)
    {
      ParameterInfo<Interval> p(Interval(left,right));
      if(right-left<1e-17)
        p.flag=uncheckable;
      c.insert(p);
    }
    left = b->c.rightBound();
    ++b;
  }

  if(left!=intv.rightBound())
  {
    ParameterInfo<Interval> p(Interval(left,intv.rightBound()));
    if(intv.rightBound()-left<1e-17)
      p.flag=uncheckable;
    c.insert(p);    
  }
}

void checkOverlapingIntervals(DataContainer& c)
{
  if(c.size()==0)
    return;
  DataContainer::iterator b=c.begin(), e=c.end(), prev;
  prev=b;
  ++b;
  Interval temp;
  while(b!=e)
  {
    if(intersection(b->c,prev->c,temp))
    {
      c.erase(prev);
    }
    prev = b;
    ++b;
  }
}

void findSuperstable(LogMap &f, DataContainer& mainContainer, Interval intv, int maxPeriod, long long grid)
{
  for(long long period=2;period<maxPeriod;++period)
  {
//    std::cout << "---------------------------\nperiod = " << period << std::endl;
    SuperstableSet superstable;
    DataContainer c;
//    std::cout << "findSuperstable " << std::endl;
    findSuperstable(superstable,period,grid,intv);
//    std::cout << "removeLowPeriods " << std::endl;
    removeLowPeriods(superstable,period);
//    std::cout << "proveSuperstable " << std::endl;
    proveSuperstable(f,c,superstable,period,intv);
//    std::cout << "computeWindowsWidth " << std::endl;
    computeWindowsWidth(f,c,period,intv);
    mainContainer.insert(c.begin(),c.end());
  }    

//  printSet(mainContainer);
//  std::cout << "checkOverlaping " << std::endl;
  checkOverlapingIntervals(mainContainer);
//  std::cout << "inserting unknown data " << std::endl;
  insertUnknownInfo(mainContainer,intv);
}

// ****************************************************************************
// procedure findSuperstable
//
// this procedure traverse the container c and calls findSuperstable procedure
// for each element in the container which has attached label 'unknown'
// ****************************************************************************
void findSuperstable(LogMap f[], DataContainer& c)
{
  int i=0;

  DataContainer::iterator p = c.begin();
  int sizeOfContainer = c.size();

  #ifdef __USE_PARALLEL__
  #pragma omp parallel shared(p,sizeOfContainer,i)
  {
    #pragma omp for schedule(dynamic,chunkSizeForParallelization)
  #endif
    for(i=0;i<sizeOfContainer;++i)
    {
      int threadNum = omp_get_thread_num();
      DataContainer::iterator current;
    #ifdef __USE_PARALLEL__
      #pragma omp critical
    #endif
      {
        current = p;
        ++p;
      } // pragma omp critical
      try{
        if(current->flag!=stability && current->flag!=periodDoubling && current->flag!=uncheckable)
        {
          DataContainer mc;
          findSuperstable(f[threadNum], mc, current->c, maximumPeriod, 2);
          #ifdef __USE_PARALLEL__
            #pragma omp critical
          #endif
          {
            if(mc.size())
            {
              c.erase(current);
              c.insert(mc.begin(),mc.end());
            }
          }
        }
      }catch(std::exception& e)
      {
        std::cout << "exception: " << e.what() << std::endl;
        exit(0);
      }
      catch(const char * s)
      {
        std::cout << "exception: " << s << std::endl;
        exit(0);
      }
    }
  #ifdef __USE_PARALLEL__
  }
  #endif
}
  


