/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Multiindex.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_VECTALG_MULTIINDEX_H_ 
#define _CAPD_VECTALG_MULTIINDEX_H_ 

#include <stdexcept>
#include <vector>
#include "capd/vectalg/Vector.h"

namespace capd{
namespace vectalg{

// Multipointer always contains nondecreasing list of indexes of variables 
// with respect to we compute partial derivatives. For example, a Multipointer mp=(0,0,2,3) 
// corresponds to partial derivative d^4/{dx_0^2 dx_2 dx_3}

class Multiindex;

class Multipointer : public Vector<int,0>
{
public:
  typedef Vector<int,0>::iterator iterator;
  typedef Vector<int,0>::const_iterator const_iterator;
  
  Multipointer(void){}
  explicit Multipointer(int A_dimension) : Vector<int,0>(A_dimension){}
  explicit Multipointer(const Multiindex&);
  Multipointer(int a_dim, int a_data[]) : Vector<int,0>(a_dim,a_data){}
  Multipointer(int a_dim,bool) : Vector<int,0>(a_dim,true){}
   
  inline int module() const{
    return dimension();
  }
  long factorial() const;
  Multipointer subMultipointer(const Multipointer& mp) const;
  
  typedef std::vector<Multipointer> MultipointersVector;
  typedef std::vector<MultipointersVector> IndicesSet;
  static const IndicesSet& generateList(int p, int k);

private:
  static std::vector<IndicesSet> indices;
  static void computeNextLevel();
  static int maxKnownLevel;   
  inline static IndicesSet& getList(int n, int k)
  {
    return indices[n*(n-1)/2+k-1];
  }
};

// -------------------------------------------------------------------------------

// For a Multiindex mi, mi[p] is a number of differentiation with respect to i-th variable.
// For example, a Multipointer mp=(0,0,2,3) in 5-dimensional space corresponds to 
// the Multiindex mi=(2,0,1,1,0).
// Hence, Multiindex agrees with standard notation and it contains an additional information 
// about the dimension of the domain of the function.

class Multiindex : public Vector<int,0>
{
public:
  typedef Vector<int,0>::iterator iterator;
  typedef Vector<int,0>::const_iterator const_iterator;
  typedef std::vector<Multiindex> MultiindexVector;
  typedef std::vector<MultiindexVector> IndicesSet;

  Multiindex(void){}
  explicit Multiindex(int A_dimension) : Vector<int,0>(A_dimension){}
  Multiindex(int A_dimension, const Multipointer&);
  Multiindex(int a_dim, int a_data[]) : Vector<int,0>(a_dim,a_data){}
  Multiindex(int a_dim,bool) : Vector<int,0>(a_dim,true){}
   
  int module() const;
  long factorial() const;
  static void generateList(int n, int k, IndicesSet& result);
};

// -------------------------------------------------------------------------------

Multipointer sumMultipointers(const Multipointer&, const Multipointer&);

}} // namespace capd::vectalg

// -------------------------------------------------------------------------------

template<typename VectorType>
typename VectorType::ScalarType power(const VectorType& v, const capd::vectalg::Multiindex& m)
{
  using namespace capd::vectalg;
  if(v.dimension()!=m.dimension())
    throw std::runtime_error("power(vector,multiindex) error: different dimensions of vector and multiindex");
  typename VectorType::ScalarType result=1;
  typename VectorType::const_iterator b=v.begin(), e=v.end();
  typename Multiindex::const_iterator i=m.begin();
  while(b!=e)
  {
    result *= power(*b,*i);
    ++b;
    ++i;
  }
  return result;
}

// -------------------------------------------------------------------------------

template<typename VectorType>
typename VectorType::ScalarType power(const VectorType& v, const capd::vectalg::Multipointer& m)
{
  using namespace capd::vectalg;
  typename VectorType::ScalarType result=1;
  typename Multipointer::const_iterator b=m.begin(), e=m.end();
  while(b!=e)
  {
    typename Multipointer::const_iterator temp=b;
    int p = *b;
    do{
      ++b;
    }while(b!=e && *b==p);
    size_t n = b-temp;
    result *= power(v[p],n);
  }
  return result;
}

#endif // _CAPD_VECTALG_MULTIINDEX_H_ 

/// @}
