#ifndef _LOG_MAP_H_
#define _LOG_MAP_H_

// ------ constants possible to change at compile time only -----------
//#define __USE_GNU_MP__ 200
//#define _DEBUG_LM_
#define __USE_PARALLEL__

const unsigned long maximumPeriod = 1025;
const int numberOfProcessors = 32;

#define chunkSizeForParallelization 1

// ---------------------------------------------------------------------

#include <set>
#include <algorithm>
#include "capd/vectalg/Vector.hpp"
#include "capd/vectalg/Matrix.hpp"
#include "ParameterInfo.h"
#include "ParameterReader.h"

#ifdef __USE_PARALLEL__
/* Test for GCC >= 4.2.0 */
  #ifdef __GNUC__
    #if __GNUC__ < 4 || \
        ( __GNUC__ == 4 && __GNUC_MINOR__ < 2 )
          #error Compiling of this program in parallel mode requires gcc-4.2 or newer. Check your compiler version: g++ -v.\
            The OpenMP libraray is also required - check the header file 'omp.h'.
    #endif
  #endif
  #include <omp.h>
#else
  #define omp_get_max_threads() 1
  #define omp_get_thread_num() 0
  #define omp_get_num_procs() 1
#endif 

#ifdef __USE_GNU_MP__
#define __HAVE_MPFR__
#include "capd/interval/MpInterval.h"
#include "capd/multiPrec/MpfrClass.h"
  typedef capd::multiPrec::MpfrClass Real;
  typedef capd::intervals::Interval<Real> Interval;
  typedef Real RoundClass;
#else
#include "capd/interval/DoubleInterval.h"
#include "capd/interval/Interval.hpp"
  typedef long double Real;
  typedef capd::rounding::DoubleRounding RoundClass;
  typedef capd::intervals::Interval<Real,RoundClass> Interval;
#endif

// ---------------------------------------------------------------------

// the following constants are readed from the config file.

extern unsigned long numberOfSkippedPoints;
extern Real accuracyForNonrigorousPeriodicOrbit;

extern Real minimalSizeOfParameterWhenProvingStableOrbit;
extern Real minimalSizeOfParameterWhenExtendingStableOrbit;
extern Real minimalSizeOfParameterWhenApproachPeriodDoubling;
extern Real sizeOfSetForTheNewtonOperator;
extern Real sizeOfSetForTheNewtonOperatorWhenFillGaps;
extern Real sizeOfSetForTheNewtonOperatorWhenApproachPeriodDoubling;
extern Real sizeOfSetForTheNewtonOperatorWhenProvingStableOrbitForMinimalParameterDiam;

typedef capd::vectalg::Vector<Interval,0> IVector;
typedef std::set< ParameterInfo<Interval> > DataContainer;
typedef capd::vectalg::RowVector<Interval,0> RowVector;

struct Derivatives{
  Derivatives(const Interval& _x, const Interval& _c)
    : x(_x), dx(1.), d2xx(0.), d3xxx(0.), c(_c), dc(0.), d2cx(0.)
  {} 
  Interval x, dx, d2xx, d3xxx, c, dc, d2cx;
};

struct RealDerivatives{
  RealDerivatives(const Real& _x, const Real& _c)
    : x(_x), dx(1.), d2xx(0.), d3xxx(0.), c(_c), dc(0.), d2cx(0.)
  {} 
  Real x, dx, d2xx, d3xxx, c, dc, d2cx;
};

#include "LogisticMap.h"
typedef  LogisticMap<Interval> LogMap;

// in file DataContainer.cpp
void createFirstGrid(DataContainer& c, int n, Interval p);
void saveDataContainer(const char filename[], DataContainer& c, bool clearFile);
unsigned long readDataFromFile(const char filename[], DataContainer& c, int fromLine, unsigned long maxNumberOfLines);
void computeTotalWidth(const char filename[], bool separatePeriods=false);
void releaseMemory(DataContainer& c);
void collapseData(DataContainer& c);


// in file NewtonOperator.cpp
bool checkDisjointSets(const Interval periodicTrajectory[], unsigned long period);
bool computeNewtonOperator(
    LogMap& f, 
    unsigned long period,
    const Interval* center,
    const Interval* X,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& product
  );
ParameterFlag verifyStablePeriodicOrbit(
    LogMap& f, 
    unsigned long period,
    const Interval* center,
    const Interval* X,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& o_product
  );  
ParameterFlag verifyStablePeriodicOrbit(
    LogMap& f, 
    unsigned long period,
    Real diameter,
    Real* periodicPoint,
    int threadNumber,
    Interval* o_periodicTrajectory,
    Interval& o_product,
    bool useBrouwer
  );

// in file KrawczykOperator.cpp
void computeKrawczykOperator(
    LogMap& f,
    const Interval* center,
    const Interval* X,
    const Interval* A,
    const Interval* fcenter,
    Interval* o_periodicTrajectory,
    unsigned long period,
    int threadNumber
   );

// in file nonrigorousAlgorithms.cpp
Real computeNewtonOperator(
    LogMap& f, 
    Real x,
    unsigned long period,
    Real* o_periodicTrajectory,
    int numberOfIterations,
    int threadNumber
  );
unsigned long searchStablePeriodicOrbit(LogMap& f, Real* o_periodicPoint);

// in file StableOrbit.cpp
Real estimateSizeOfSetForIntervalNewtonOperator(Real periodicPoint[], unsigned long period, bool minimalSize);
void verifyStablePeriodicOrbit(
    LogMap& f, 
    ParameterInfo<Interval>& p, 
    const ParameterInfo<Interval>& prev,
    int threadNumber,
    unsigned long period,
    bool useBrouwer
  );

// in file Subdivide.cpp
void subdivideContainer(LogMap f[], DataContainer& c, int numberOfThreads);
void scanForPeriodDoubling(LogMap logMap[], DataContainer& c);
void fillGaps(LogMap logMap[], DataContainer& c);

//in file PeriodDoubling.cpp
bool verifyPeriodDoubling(
    LogMap& f, 
    DataContainer& c,
    int threadNumber
  );

// in file findSuperstable.cpp
void findSuperstable(LogMap &f, DataContainer& mainContainer, Interval intv, int maxPeriod, long long grid);
void findSuperstable(LogMap f[], DataContainer& c);

#endif

