#include "logMap.h"
#include "backwardShooting.h"

//*******************************************************************************************
// algorithm computeNewtonOperator
//
// input: f - logistic map
//        x - approximate periodic point
//        period - expected period
//
// output: one step nonrigorous Newton method for the function (f(x1)-x2,f(x2)-x3,...,f(xn)-x1)
//
// assumptions - class MapType must provide:
//  - public type FloatType 
//  - operator() computing value of the function
//  - operator[] computing derivative of the function
//*******************************************************************************************
Real computeNewtonOperator(
    LogMap& f, 
    Real x,
    unsigned long period,
    Real* o_periodicTrajectory,
    int numberOfIterations,
    int threadNumber
  )
{
  static Real da[numberOfProcessors*maximumPeriod],
               dcenter[numberOfProcessors*maximumPeriod],
               dg[numberOfProcessors*maximumPeriod];

  Real *a = da + threadNumber*maximumPeriod,
       *center = dcenter + threadNumber*maximumPeriod,
       *g = dg + threadNumber*maximumPeriod;

  RoundClass::roundNearest();

  // we store in the center table an approximate nonrigorous trajectory
  unsigned long i;
  center[0] = x;
  a[0] = f[x];
  for(i=1;i<period;++i)
  {
    center[i] = f(center[i-1]);
  }
  
  Real product=0;
  for(int j=1;j<=numberOfIterations;++j)
  {
    unsigned long i;
  
    // having the center and the set we can compute F and F'
    for(i=0;i<period;++i)
    {
      g[i] = center[(i+1)%period] - f(center[i]);
      a[i] = f[center[i]];
    }
    
    int info = backwardShooting(period,a,center,g,o_periodicTrajectory,product);
    if( info%2 == 0 )
    {
    // if we cannot use backward shooting, try to solve the system using Gauss elimination
      capd::vectalg::Matrix<Real,0,0> M(period,period);
      for(i=1;i<=period;++i)
      {
        M(i,i) = -a[i-1];
        M(i,(i%period)+1) += 1.; // += for the case period=1
      }
      capd::matrixAlgorithms::gauss(M,g,o_periodicTrajectory);
    }
  
    // the real Newton operator
    Real* target = (j==numberOfIterations) ? o_periodicTrajectory : center;
    for(i=0;i<period;++i)
    {
      target[i] = center[i] - o_periodicTrajectory[i];
    }
  }
  return product;
}

//*******************************************************************************************
// algorithm searchStablePeriodicOrbit
// input: f - a logistic map
//        numberOfSkippedPoints - we start from a critical point and skip first points on its trajectory
//        accuracy - of periodic point
// 
// output: period of approximate periodic point, if found
//         approximate periodicPoint
//
// assumptions - class MapType must provide:
//  - public type FloatType 
//  - method criticalPoint()
//  - operator() computing value of the function
//*******************************************************************************************
unsigned long searchStablePeriodicOrbit(
    LogMap& f, 
    Real* o_periodicPoint
  )
{
  RoundClass::roundNearest();

  // first we take a critical point and compute its trajectory
  Real x = f.criticalPoint();
  // skip some points
  for(unsigned long i=0;i< numberOfSkippedPoints;++i){
    x = f(x);
  }
  // next we look if the trajectory looks like periodic orbit
  o_periodicPoint[0] = x;
  unsigned long period;
  Real product = 1;
  for(period=1;period<maximumPeriod;++period)
  {
    product *= f[o_periodicPoint[period-1]];
    o_periodicPoint[period] = f(o_periodicPoint[period-1]);
    if( capd::abs(o_periodicPoint[period]-o_periodicPoint[0]) < accuracyForNonrigorousPeriodicOrbit ) 
      break;
  }

  // if the loop finishes with period==maximumPeriod then we did not find good candidate
  // also if an orbit seems to be unstable
  if (period==maximumPeriod || capd::abs(product)>Real(1.))
    return 0;

  static std::set<int> maximalPeriodFound;
  #pragma omp critical
  {
    std::set<int>::iterator q = maximalPeriodFound.find(period);
    if(q==maximalPeriodFound.end())
    {
      std::cout << "possible period found: " << period << ", c=" << f.midC << "\ndetected periods: { ";
      maximalPeriodFound.insert(period);
      q=maximalPeriodFound.begin();
      while(q!=maximalPeriodFound.end())
      {
        std::cout << *q << " ";
        ++q;
      }
      std::cout << "}" << std::endl;
    }
  }
  return period;
}

