#include "logMap.h"


//*******************************************************************************************
// algorithm verifyCMax
// input: f - a logistic map
// 
//*******************************************************************************************
bool verifyCMax(
    LogMap& f,
    const IVector& xRange, 
    const Interval& c, 
    Real rightPeriodicPoint[], 
    unsigned long period, 
    int threadNumber,
    Interval* periodicTrajectory
  )
{
  Interval product;// temporary argument for general algorithm

  f.setParameter(c.right());
  IVector center(period);
  unsigned long i;
  for(i=0;i<period;++i)
    if(!xRange[i].contains(rightPeriodicPoint[i]))
    {
      center[i] = xRange[i].mid();
    }else
      center[i]=rightPeriodicPoint[i];

  bool exists = computeNewtonOperator(f,period,center.begin(),xRange.begin(),threadNumber,periodicTrajectory,product);
  if(!exists)
  {
//    std::cout << "nie udalo sie sprawdzic istnienia dla cMax, diam(xRange)=" << capd::vectalg::size(xRange)<< std::endl;
    return false;
  }

  for(i=0;i<period;++i)
    if(!periodicTrajectory[i].subset(xRange[i]))
    {
//      std::cout << "istnienie orbity dla cMax poza xRange" << std::endl;      
      return false;
    }
  // we have to verify d/dx G(cMax,x(xMax))<0
  // G(xMin,cMax) < 0
  // G(xMax,xMax) > 0
  Interval dx = f[periodicTrajectory[0]];
  Interval GxMin = f(xRange[0].left());
  Interval GxMax = f(xRange[0].right());
  
  for(i=1;i<2*period;++i)
  {
    // derivative at cMax on the fixed point curve must be smaller than -1
    if(i==period && !(dx<-1.))
    {
//      std::cout << "cMax - pochdna nie jest mniejsza od -1" << std::endl;
      return false;
    }
    dx = f[periodicTrajectory[i%period]]*dx;
    GxMin = f(GxMin);
    GxMax = f(GxMax);
  }
  dx = 1.-dx;
  GxMin = xRange[0].left() - GxMin;
  GxMax = xRange[0].right() - GxMax;
  
#ifdef _DEBUG_LM_
  if (!( (dx.rightBound()<0) && (GxMin.rightBound()<0) && (GxMax.leftBound()>0) ))
  {
    std::cout << "dx:" << dx << std::endl;
    std::cout << "GxMin:" << GxMin << std::endl;
    std::cout << "GxMax:" << GxMax << std::endl << std::endl;
  }
#endif
  RoundClass::roundNearest();
  return ( (dx.rightBound()<0) && (GxMin.rightBound()<0) && (GxMax.leftBound()>0) );
}


//*******************************************************************************************
// algorithm verifyCMin
// input: f - a logistic map
// 
//*******************************************************************************************
bool verifyCMin(LogMap& f,const Interval& x, const Interval& c, unsigned long period)
{
  int i;
  f.setParameter(c.left());
  const int size=1000;
  Real grid[size+1];
  Real step = (x.rightBound()-x.leftBound())/size;
  grid[0] = x.leftBound();
  for(i=1;i<size;++i)
    grid[i] = grid[i-1] + step;
  grid[size] = x.rightBound();

  for(i=0;i<size;++i)
  {
    Interval u(grid[i],grid[i+1]);
    Interval dx(1);
    for(unsigned long j=0;j<2*period;++j)
    {
      dx = f[u]*dx;
      u = f(u);
    }
    if (! ((Real(1)-dx.leftBound()) >0 ) )
    {
//        std::cout << "dx(cmin): " << dx << std::endl;
      return false;
    }
  }
    
  return true;
}

//*******************************************************************************************
// algorithm verifyC3Condition
// input: f - a logistic map
// 
//*******************************************************************************************
bool verifyC3Condition(LogMap& f, IVector& xRange, const Interval& cRange, unsigned long period,int threadNumber)
{
  RoundClass::roundNearest();
  const int sizeX = 100;
  const int sizeC = 100;
  Real gridX[sizeX+1];
  Real gridC[sizeC+1];
  Real stepX = (xRange[0].rightBound()-xRange[0].leftBound())/sizeX;
  Real stepC = (cRange.rightBound()-cRange.leftBound())/sizeC;

  gridX[0] = xRange[0].leftBound();
  gridC[0] = cRange.leftBound();

  int i;
  for(i=1;i<sizeX;++i)
    gridX[i] = gridX[i-1] + stepX;
  for(i=1;i<sizeC;++i)
    gridC[i] = gridC[i-1] + stepC;
  gridX[sizeX] = xRange[0].rightBound();
  gridC[sizeC] = cRange.rightBound();

  Interval d2cx,d2xx,dxdc;

  // computation of x'(c)
  bool first=true;
  IVector center(period);
  capd::vectalg::mid(xRange,center);
  Interval product;
  static Interval dxTrajectory[numberOfProcessors*maximumPeriod];
  static Real  dPeriodicPoint[numberOfProcessors*maximumPeriod];
  Interval* traj = dxTrajectory + threadNumber*maximumPeriod;
  Real* perPoint = dPeriodicPoint + threadNumber*maximumPeriod;
  IVector c(period);

  for(i=0;i<sizeC;++i)
  {    
    f.setParameter(Interval(gridC[i],gridC[i+1]));
    // check if a curve of fixed point for f^period exists
    RoundClass::roundNearest();
    computeNewtonOperator(f,center[0].mid().leftBound(), period,perPoint,3,threadNumber);
    std::copy(perPoint,perPoint+period,c.begin());
    if(!subset(c,xRange))
    {
      c=center;
    }
    if( ! computeNewtonOperator(f,period,c.begin(),xRange.begin(),threadNumber,traj,product) )
    {
//      std::cout << "x(c) moze nie istniec, diam=" << capd::vectalg::size(xRange) << std::endl;
      return false;
    }
    // check if the map is injective 
    if(isSingular(product))
    {
//      std::cout << "brak injektywnosci" << std::endl;
      return false;
    }
    // check if 'period' is really minimal period
    if(!checkDisjointSets(traj,period))
    {
//      std::cout << "x(c) = zbior nie jest rozlaczny" << std::endl;
      return false;
    }

    Interval dx(1), dc(0);

    for(unsigned long j=0;j<period;++j)
      f.composeDerivatives(traj[j],f.getParameter(),dx,dc);

    if(first)
    {
      dxdc = dc/(1.-dx);
      first = false;
    } else
      dxdc = capd::intervals::intervalHull(dxdc,dc/(1.-dx));

    // check if x(c) is monotonic
    if(isSingular(dxdc))
    {
//          std::cout << "x'(c) = 0" << std::endl;
      return false;
    }
  }
  
  // computation od d2cx, d2xx, d3xxx
  first=true;

  for(i=0;i<sizeC;++i)
  {  
    for(int j=0;j<sizeX;++j)
    {
      Derivatives der(Interval(gridX[j],gridX[j+1]),Interval(gridC[i],gridC[i+1]));
      for(unsigned long j=0;j<2*period;++j)
        f.composeDerivatives(der);
      
      if(first)
      {
        d2cx = -der.d2cx;
        d2xx = -der.d2xx;
        first = false;
      }
      else
      {
        d2cx = capd::intervals::intervalHull(d2cx,-der.d2cx);
        d2xx = capd::intervals::intervalHull(d2xx,-der.d2xx);
      }
        
      if (der.d3xxx.rightBound()>=0)
      {
//          std::cout << "d3xxx = " << der.d3xxx << std::endl;
        return false;
      }
    }
  }

/*
  if(! ( (d2cx + d2xx*dxdc*Interval(0,1)).rightBound()<0. ))
  {
    std::cout << "d2cx:" << d2cx << std::endl;
    std::cout << "d2xx:" << d2xx << std::endl;
    std::cout << "dxdc:" << dxdc << std::endl;
    std::cout << "last:" << (d2cx + d2xx*dxdc*Interval(0,1)) << std::endl;
  }
*/
  return  (d2cx + d2xx*dxdc*Interval(0,1)).rightBound()<0.; 
}

//*******************************************************************************************
// algorithm verifyPeriodDoubling
// input: f - a logistic map
//        beforeBifurcation - a parameter info on periodic orbit before possible bifurcation
//        afterBifurcation - a parameter info on periodic orbit after possible bifurcation
//        c - data container which contains information about all computed subintervals
//
// output: modufied data container c if the period doubling is sucessfully verified
//
// assumptions - class MapType must provide:
//  - public type IntervalType
//  - public type FloatType
//*******************************************************************************************
bool verifyPeriodDoubling(
    LogMap& f, 
    DataContainer& c,
    int threadNumber
  )
{
  DataContainer::iterator beforeBifurcation = c.begin();
  DataContainer::iterator afterBifurcation = --(c.end());
  static Real dPeriodicPoint[maximumPeriod*numberOfProcessors];
  Real* periodicPoint = dPeriodicPoint + threadNumber*maximumPeriod;


  // first we define a set in the 2D space of the form [x1,x2]\times [c1,c2]
  // on which period doubling will be verified
  // [c1,c2] is easily a gap between intervals defined in beforeBifurcation and afterBifurcation
  // interval [x1,x2] must be check more carefully, since rigorous estimations for stable periodic
  // orbits before and after bifurcation might be shifted. 

  unsigned long period = beforeBifurcation->enclosure->dimension();
  Interval cRange (beforeBifurcation->c.rightBound(),afterBifurcation->c.leftBound());
  IVector xRange(period), center(period), traj1(2*period), traj2(2*period);
  Interval temp;

  ParameterInfo<Interval> after = *afterBifurcation;
  ParameterInfo<Interval> before = *beforeBifurcation;


// we start to modify the range - to be as close as possible from both sides to the bifurcation parameter
//  std::cout << "rozpoczynam modyfikacje, okres=" << period << ", c=" << cRange << std::endl;
  DataContainer::iterator counter=beforeBifurcation;
  int numberOfElements = 0;
  while(counter!=afterBifurcation)
  {
    ++numberOfElements;
    ++counter;
  }
  counter=beforeBifurcation;
  ++counter;
  c.erase(counter,afterBifurcation);

  Real diam = 1e-10;
  IVector* dest = &center;
  IVector& enc = *(before.enclosure);
  bool first=true;
  Real size = sizeOfSetForTheNewtonOperatorWhenApproachPeriodDoubling;
  do{
    Real leftEnd = cRange.leftBound();
    Real rightEnd = leftEnd + diam;
    f.setParameter(Interval(leftEnd,rightEnd));
    computeNewtonOperator(f,enc[0].mid().leftBound(), period,periodicPoint,4,threadNumber);
    ParameterFlag flag = verifyStablePeriodicOrbit(f,period,size,periodicPoint,threadNumber,dest->begin(),temp,false);
    if(flag==stability)
    {
      if(!first)
        capd::vectalg::intervalHull(xRange,center,*dest);
      if(!checkDisjointSets(dest->begin(),period))
        break;
      
      first = false;
      cRange.setLeftBound(rightEnd);
      if (dest == &center)
        dest = &xRange;
      else
        dest = &center;
      
      if(! (temp<-0.7)) // to far from bif point
        break;
    }else if(diam>minimalSizeOfParameterWhenApproachPeriodDoubling)
    {
      diam /= 1.2;
//      std::cout << "lewa strona podzielilem diam=" << diam << ", c=" << cRange << std::endl;
    }
    else
      break;
  }while(true);

  if(dest== &center)
    dest = &xRange;
  else
    dest = &center;

  if(!first)
  {
    ParameterInfo<Interval> p(
      Interval(before.c.rightBound(),cRange.leftBound()),
      stability,dest->begin(),period
      );
    c.insert(p);
    beforeBifurcation = c.find(p);
  }

  // now we try to extend the stability range after bifurcation as close as possible
  // to the bifurcation parameter 
  first=true;
  dest = &traj1;
  diam = 1e-10;
  f.setParameter(cRange.right());
  computeNewtonOperator(f,after.enclosure->begin()->mid().leftBound(), 2*period,periodicPoint,10,threadNumber);
  do{
    Real rightEnd = cRange.rightBound();
    Real leftEnd = rightEnd - diam;
    f.setParameter(Interval(leftEnd,rightEnd));
    computeNewtonOperator(f,periodicPoint[0], 2*period,periodicPoint,3,threadNumber);
    ParameterFlag flag = verifyStablePeriodicOrbit(f,2*period,size,periodicPoint,threadNumber,dest->begin(),temp,false);
    if(flag==stability)
    {
      if(!first)
        capd::vectalg::intervalHull(traj1,traj2,*dest);
      if( !checkDisjointSets(dest->begin(),2*period))
        break;

      first = false;
      if(dest=&traj1)
        dest = &traj2;
      else
        dest = &traj1;
      cRange.setRightBound(leftEnd);        

      if( !(temp>0.7)) // to far from bif point
      {
//        std::cout << "pochodna z prawej = " << temp << std::endl;
        break;
      }
    }else if(diam>minimalSizeOfParameterWhenApproachPeriodDoubling)
    {
      diam /= 1.2;
//      std::cout << "prawa strona podzielilem diam=" << diam << ", c=" << cRange << std::endl;
    }
    else
      break;
  }while(true);
  

  if(dest== &traj1)
    dest = &traj2;
  else
    dest = &traj1;
  
//  if(first)
//    std::cout << "nie przyblizylem z prawej\n";

  if(!first)
  {
    ParameterInfo<Interval> p(
      Interval(cRange.rightBound(),after.c.leftBound()),
      stability,dest->begin(),2*period
      );
    c.insert(p);
    afterBifurcation = c.find(p);
  }

  createFirstGrid(c,numberOfElements,cRange);
//  std::cout << "zakonczylem modyfikacje, okres=" << period << ", c=" << cRange << std::endl;
//  exit(0);  

  // we try to obtain better enclosure after bifurcation
  f.setParameter(cRange.right());
  computeNewtonOperator(f,after.enclosure->begin()->mid().leftBound(), 2*period,periodicPoint,10,threadNumber);

  for(unsigned long i=0;i<period;++i)
  {
    Interval temp(
      capd::min(periodicPoint[i],periodicPoint[i+period]),
      capd::max(periodicPoint[i],periodicPoint[i+period])
      );
    xRange[i] = temp+Interval(-1,1)*1e-11;
  }
  computeNewtonOperator(f,after.enclosure->begin()->mid().leftBound(), period,periodicPoint,10,threadNumber);

  // the existence of period doubling is verified in the following steps  
//  std::cout << "period:" << period << std::endl;
  if( ! verifyCMax(f,xRange ,cRange,periodicPoint,period,threadNumber,center.begin()) )
  {
//    std::cout << "cMax not verified, period=" << period << std::endl;
    return false;
  }
  if( ! verifyCMin(f,xRange[0],cRange,period) )
  {
//    std::cout << "cMin not verified, period=" << period << std::endl;
    return false;
  }

  if( ! verifyC3Condition(f,xRange,cRange,period,threadNumber) )  
  {
//    std::cout << "last condition not verified, period=" << period << std::endl;
    return false;
  }

  c.erase(++beforeBifurcation,afterBifurcation);
  c.insert(ParameterInfo<Interval>(cRange,periodDoubling,xRange.begin(),period));
  return true;
}

