#include "logMap.h"

Real estimateSizeOfSetForIntervalNewtonOperator(Real periodicPoint[], unsigned long period, bool fillGaps)
{
  RoundClass::roundNearest();
  if (fillGaps)
    return sizeOfSetForTheNewtonOperatorWhenFillGaps;
  return sizeOfSetForTheNewtonOperator/(1+log10(period));
}


// ****************************************************************************
// algorithm verifyStablePeriodicOrbit
// 
// input: f - a logistic map
//        p - ParameterInfo, interval to be verified
//        prev - previous interval of parameters
//               if it contains stability information that we can expect that
//               the curve of periodic orbit will continue. This speed up computations
//
// output: a flag in data structure p is modified. If stability verified, an enclosure 
//         for thsis orbit is also stored in structure p            
// ****************************************************************************
void verifyStablePeriodicOrbit(
    LogMap& f, 
    ParameterInfo<Interval>& p, 
    const ParameterInfo<Interval>& prev,
    int threadNumber,
    unsigned long period,
    bool useBrouwer
  )
{
  bool fillingGaps = (period!=0);
  Interval product; // temporary argument for general algorithm
  static Real dPeriodicPoint[numberOfProcessors*maximumPeriod];
  static Interval dPeriodicTrajectory[numberOfProcessors*maximumPeriod];  

  Real *periodicPoint = dPeriodicPoint + threadNumber*maximumPeriod;
  Interval* periodicTrajectory = dPeriodicTrajectory + threadNumber*maximumPeriod;

  unsigned long i,j;
  f.setParameter(p.c);
  if( !period )
  {
    period = searchStablePeriodicOrbit(f,periodicPoint);
    if(!period) // cannot find periodic orbit
    {
      p.flag = unknown;
      return;
    }
    //try to improve an orbit by means of the real Newton operator
    computeNewtonOperator(f,periodicPoint[0],period,periodicPoint,3,threadNumber);
  }
  else
  {
    period = prev.enclosure->dimension();
    computeNewtonOperator(f,(*(prev.enclosure))[0].mid().leftBound(), period,periodicPoint,10,threadNumber);
  }

  Real diameter = estimateSizeOfSetForIntervalNewtonOperator(periodicPoint,period,fillingGaps);
  p.flag = verifyStablePeriodicOrbit(f,period,diameter,periodicPoint,threadNumber,periodicTrajectory,product,useBrouwer);
  
  // if proof succeeded, we save rigorous bound of the trajectory in data container
  if(p.flag==stability){
    p.enclosure = new IVector(period,periodicTrajectory);
//    std::cout << "sprawdzilem na calym, okres=" << period << ", pochodna=" << product << ", c=" << f.getParameter() << std::endl;
//    exit(0);
    return;
  }
  Real paramSize = minimalSizeOfParameterWhenProvingStableOrbit;
  if(paramSize > (p.c.rightBound()-p.c.leftBound()))
  {
    p.flag = uncheckable;
    return;
  }
  
  Real midPoint = p.c.mid().leftBound();
  Real leftEnd  = capd::max(Real(p.c.leftBound()),  midPoint - paramSize);
  Real rightEnd = capd::min(Real(p.c.rightBound()), midPoint + paramSize); 
  f.setParameter(Interval(leftEnd,rightEnd));
  p.flag = verifyStablePeriodicOrbit(
      f,period,
      sizeOfSetForTheNewtonOperatorWhenProvingStableOrbitForMinimalParameterDiam,periodicPoint,
      threadNumber,periodicTrajectory,product,useBrouwer
    );

  if(p.flag!=stability)
  {
//    std::cout << "nie sprawdzilem w srodku, okres=" << period <<  ", c=" << f.getParameter() << std::endl;
    return;
  }

  p.enclosure = new IVector(period,periodicTrajectory);
  IVector& enc = *(p.enclosure);

  if(p.c==f.getParameter())
    return;

  Interval parameterRange = f.getParameter();
  // to prevent unnecessary copying of data we just swap pointers
  Interval *destination = periodicTrajectory;

  // try to extend this interval to the right and left sides
  Real leftPeriodicPoint = periodicPoint[0];
  Real rightPeriodicPoint = periodicPoint[0];

  const static Real factor = 1.2;
//  std::cout << "rozpoczynam rozszerzanie:" << capd::intervals::diam(parameterRange) << ", okres=" << period << std::endl;

  Real diam = (p.c.rightBound()-p.c.leftBound())/4;
  
  leftEnd = capd::min(p.c.rightBound(),parameterRange.rightBound());
  rightEnd = capd::min(Real(p.c.rightBound()),leftEnd+diam);
  while(leftEnd<p.c.rightBound())
  {
    f.setParameter(Interval(leftEnd,rightEnd));
    // try to improve a little bit
    computeNewtonOperator(f,rightPeriodicPoint,period,periodicPoint,4,threadNumber);
    // check stability
    diameter = estimateSizeOfSetForIntervalNewtonOperator(periodicPoint,period,fillingGaps);
    ParameterFlag flag = verifyStablePeriodicOrbit(f,period,diameter,periodicPoint,threadNumber,destination,product,useBrouwer);
    if(flag!=stability)
    {
      diam /= factor;
//      std::cout << "right side diam = " << diam << std::endl;
      if(diam < minimalSizeOfParameterWhenExtendingStableOrbit)
        break;
      rightEnd = capd::min(Real(p.c.rightBound()),leftEnd+diam);
      continue;
    }

    // if done take an interval enclosure for both parts, to don't store two intervals in container
    for(j=0;j<period;++j)
      destination[j] = intervalHull(enc[j],periodicTrajectory[j]);
    
    if(!checkDisjointSets(destination,period))
      break;

    if(destination==periodicTrajectory)
      destination = enc.begin();
    else 
      destination = periodicTrajectory;

    if(product<-0.5)
      diam *= factor;
    parameterRange.setRightBound(rightEnd);
    leftEnd = rightEnd;
    rightEnd = capd::min(Real(p.c.rightBound()),leftEnd+diam);
    rightPeriodicPoint = periodicPoint[0];
  } // while

  diam = (p.c.rightBound()-p.c.leftBound())/4;
  rightEnd = capd::max(p.c.leftBound(),parameterRange.leftBound());
  leftEnd = capd::max(Real(p.c.leftBound()),rightEnd-diam);
  while(rightEnd>p.c.leftBound())
  {
    f.setParameter(Interval(leftEnd,rightEnd));
    // try to improve a little bit
    computeNewtonOperator(f,leftPeriodicPoint,period,periodicPoint,4,threadNumber);
    // check stability
    diameter = estimateSizeOfSetForIntervalNewtonOperator(periodicPoint,period,fillingGaps);    
    ParameterFlag flag = verifyStablePeriodicOrbit(f,period,diameter,periodicPoint,threadNumber,destination,product,useBrouwer);
    if(flag!=stability)
    {
//      std::cout << "left side diam = " << diam << std::endl;
      diam /= factor;
      if(diam < minimalSizeOfParameterWhenExtendingStableOrbit)
        break;
      leftEnd = capd::max(Real(p.c.leftBound()),rightEnd-diam);
      continue;
    }

    // if done, take an interval enclosure for both parts, to don't store two intervals in container
    for(j=0;j<period;++j)
      destination[j] = intervalHull(enc[j],periodicTrajectory[j]);
    
    if(!checkDisjointSets(destination,period))
      break;

    if(destination==periodicTrajectory)
      destination = enc.begin();
    else 
      destination = periodicTrajectory;

    parameterRange.setLeftBound(leftEnd);
    rightEnd = leftEnd;
    leftEnd = capd::max(Real(p.c.leftBound()),rightEnd-diam);
    leftPeriodicPoint = periodicPoint[0];
  } // while


  if(destination==enc.begin())
  {  
    for(j=0;j<period;++j)
      enc[j] = periodicTrajectory[j];
  }
//  std::cout << "rozszerzylem do " << capd::intervals::diam(parameterRange) << ", okres=" << period << std::endl;
  p.c = parameterRange;
}

