#include "logMap.h"

using namespace std;


// ****************************************************************************
//
// algorithm subdivideParameter
//
// input: c - data container which stores information about status 
//            of any subinterval of parameter range
//        f - logistic map
//        leftEnd - left end of subinterval to subdivide
// 
// output: - current subinterval dvided onto 2 parts with 
//            attached flags to both parts
//         - left end of the next interval to be verified
// ****************************************************************************
void subdivideParameter(
    LogMap& f, 
    DataContainer& c, 
    ParameterInfo<Interval>& p, 
    const ParameterInfo<Interval>& prev,
    int threadNumber,
    unsigned long expectedPeriod,
    bool useBrouwer
  )
{  
  if( p.flag==stability || p.flag==periodDoubling || p.flag==uncheckable)
  {
    c.insert(p);
    return;
  }

  Real leftEnd = p.c.leftBound();
  Real rightEnd = p.c.rightBound();

// only approximate value is enough
  RoundClass::roundNearest();
  Real midPoint = (leftEnd+rightEnd)/2;
  
  // if the interval is to small to subdivide it
  // then take whole subinterval
  if(midPoint<=leftEnd || midPoint>=rightEnd) 
  {
    verifyStablePeriodicOrbit(f,p,prev,threadNumber,expectedPeriod,useBrouwer);
    if(leftEnd<p.c.leftBound())
    {
      ParameterInfo<Interval> extraNodeLeft(Interval(leftEnd,p.c.leftBound()));
      c.insert(extraNodeLeft);
    }
    if(rightEnd>p.c.rightBound())
    {
      ParameterInfo<Interval> extraNodeRight(Interval(p.c.rightBound(),rightEnd));
      c.insert(extraNodeRight);
    }
  // finally, we add new node to a data structure
    c.insert(p);
    return;
  }

// next we insert two new nodes to the set
  ParameterInfo<Interval> left(Interval(leftEnd,midPoint));
  ParameterInfo<Interval> right(Interval(midPoint,rightEnd));

// and verify stability on left subinterval
  verifyStablePeriodicOrbit(f,left,prev,threadNumber,expectedPeriod,useBrouwer);
  // this is in case when high period orbit has been found 
  // on an interval smaller than this stored previously in left
  if(leftEnd<left.c.leftBound())
  {
    ParameterInfo<Interval> extraNodeLeft(Interval(leftEnd,left.c.leftBound()));
    c.insert(extraNodeLeft);
  }
  if(midPoint>left.c.rightBound())
  {
    ParameterInfo<Interval> extraNodeRight(Interval(left.c.rightBound(),midPoint));
    c.insert(extraNodeRight);
  }
// finally, we add new nodes to a data structure
  c.insert(left);

// the same on right side, now left or extraNodeRight is the last computed interval
// however, extraNodeRight does not contain stability information
// If left does not contains stability information 
// then we use as a previous prev
  const ParameterInfo<Interval>& prev2 = (left.flag==stability) ? left : prev;
  verifyStablePeriodicOrbit(f,right,prev2,threadNumber,expectedPeriod,useBrouwer);
  if(midPoint<right.c.leftBound())
  {
    ParameterInfo<Interval> extraNodeLeft(Interval(midPoint,right.c.leftBound()));
    c.insert(extraNodeLeft);
  }
  if(rightEnd>right.c.rightBound())
  {
    ParameterInfo<Interval> extraNodeRight(Interval(right.c.rightBound(),rightEnd));
    c.insert(extraNodeRight);
  }
  c.insert(right);
}



// ****************************************************************************
// procedure subdivideContainer
//
// this procedure traverse the container c and calls subdivideParameter algorithm
// for each element in the container
// the elements which has been previously computed are skipped
// ****************************************************************************
void subdivideContainer(LogMap f[], DataContainer& c, int numberOfThreads)
{
// temporary element. We need some argument which is used as a previous to 
// the first element in the container
  ParameterInfo<Interval> prev[numberOfProcessors];
  int i=0;

  DataContainer data[numberOfProcessors];
  DataContainer::iterator p = c.begin();
  int sizeOfContainer = c.size();

  #ifdef __USE_PARALLEL__
  #pragma omp parallel shared(p,data,sizeOfContainer,prev,i)
  {
    #pragma omp for schedule(dynamic,chunkSizeForParallelization)
  #endif
    for(i=0;i<sizeOfContainer;++i)
    {
      int threadNum = omp_get_thread_num();
      ParameterInfo<Interval> current;
    #ifdef __USE_PARALLEL__
      #pragma omp critical
    #endif
      {
        current = *p;
        ++p;
      } // pragma omp critical
      try{
//        std::cout << "processing element=" << i << std::endl;
        subdivideParameter(f[threadNum],data[threadNum],current,prev[threadNum],threadNum,0,true);
      }catch(std::exception& e)
      {
        std::cout << "exception: " << e.what() << std::endl;
        exit(0);
      }
      catch(const char * s)
      {
        std::cout << "exception: " << s << std::endl;
        exit(0);
      }
    }
  #ifdef __USE_PARALLEL__
  }
  #endif
  c.clear();
  for(i=0;i<numberOfProcessors;++i)
    c.insert(data[i].begin(),data[i].end());
}


// ****************************************************************************
// procedure subdivideContainer - version for one thread, used in the function
// 'scanForGaps'
//
// this procedure traverse the container c and calls subdivideParameter algorithm
// for each element in the container
// the elements which has been previously computed are skipped
// this algorithm runs in one thread and is designed for filling te gaps
// in stable period windows. 
// ****************************************************************************

void subdivideContainer(LogMap& f, DataContainer& c, int threadNum)
{
  ParameterInfo<Interval> first = *(c.begin());
  DataContainer::iterator p = c.begin();

  int sizeOfContainer = c.size();
  int expectedPeriod = first.enclosure->dimension();
  DataContainer result;

  for(int i=0;i<sizeOfContainer;++i)
  {
      ParameterInfo<Interval> current = *p;
      DataContainer::iterator temp = p;
      ++p;
      subdivideParameter(f,result,current,first,threadNum,expectedPeriod,true);
  }
  c.clear();
  c.insert(result.begin(),result.end());
}

// ****************************************************************************
// procedure scanForPeriodDoubling
//
// this procedure traverse the container c and search for possible period doubling 
// or period halfing. If such a possibility is found then a general procedure 
// for proving period doubling is called
// ****************************************************************************
void scanForPeriodDoubling(LogMap logMap[], DataContainer& c)
{
  std::vector<DataContainer> perDoubling;
  DataContainer::iterator b = c.begin(), e = c.end();

  while(b!=e)
  {
    if(b->flag!=stability)
    {
      ++b;
      continue;
    }
    DataContainer::iterator next = b;
    ++next;
    while(next!=e && next->flag==stability)
    {
      ++next;
      ++b;
    }
    
    // we reach the end of container
    if(next==e)
      break;
    // period doubling already proven 
    if(next->flag==periodDoubling)
    {
      ++b;
      continue;
    }

    DataContainer::iterator beforePossiblePeriodDoubling = b;
    do{
      ++b;
    }while(b!=e && b->flag!=stability);

    if(b==e)
      break;
    if(2*beforePossiblePeriodDoubling->enclosure->dimension() == b->enclosure->dimension())
    {
      DataContainer pd(beforePossiblePeriodDoubling,++b);
      perDoubling.push_back(pd);
      c.erase(beforePossiblePeriodDoubling,b);
    }
  }

  int j = perDoubling.size();
  cout << "(" << j << ") ..." << endl;
  int i;
  #ifdef __USE_PARALLEL__
  #pragma omp parallel shared(perDoubling,logMap,i)
  {
    #pragma omp for schedule(dynamic)
  #endif
    for(i=0;i<j;++i)
    {
      #pragma omp critical
      {
      std::cout << "i=" << i+1 << ", " 
                << "possible period doubling:" << perDoubling[i].begin()->enclosure->dimension() 
                << "->" << perDoubling[i].rbegin()->enclosure->dimension()
                << ", c=" << Interval(perDoubling[i].begin()->c.rightBound(),perDoubling[i].rbegin()->c.leftBound()) 
                << std::endl;
      }
      
      try{
        int threadNum = omp_get_thread_num();
        verifyPeriodDoubling(logMap[threadNum], perDoubling[i],threadNum);
      }catch(std::exception& e)
      {
        std::cout << "exception: " << e.what() << std::endl;
        exit(0);
      }
      catch(const char * s)
      {
        std::cout << "exception: " << s << std::endl;
        exit(0);
      }
    }
  #ifdef __USE_PARALLEL__
  }
  #endif

  for(i=0;i<j;++i)
    c.insert(perDoubling[i].begin(),perDoubling[i].end());
}


// ****************************************************************************
// procedure fillGaps
//
// this procedure traverse the container c and search for possible gaps 
// in the stability range of the same period
// ****************************************************************************
void fillGaps(LogMap logMap[], DataContainer& c)
{
  DataContainer::iterator b = c.begin(), e = c.end();
  std::vector<DataContainer> gaps;

  while(b!=e)
  {
    if(b->flag!=stability)
    {
      ++b;
      continue;
    }

    DataContainer::iterator next = b;
    ++next;
    while(next!=e && next->flag==stability)
    {
      ++next;
      ++b;
    }
    
    // we reach the end of container
    if(next==e)
      break;
    
    if(next->flag==periodDoubling)
    {
      ++b;
      continue;
    }
    DataContainer::iterator beforePossibleGap = b;
    int j=0;
    do{
      ++b;
      ++j;
    }while(b!=e && b->flag!=stability);

    if(b==e)
      break;

    if(beforePossibleGap->enclosure->dimension() == b->enclosure->dimension())
    {
      DataContainer gap;
      
      createFirstGrid(gap,j/2,Interval(beforePossibleGap->c.rightBound(),b->c.leftBound()));
      gap.insert(*beforePossibleGap);
      gaps.push_back(gap);
      c.erase(beforePossibleGap,b);
    }
  }

  int j = gaps.size();
  cout << "(" << j << ") ..." << endl;
  int i;
  #ifdef __USE_PARALLEL__
  #pragma omp parallel shared(gaps,logMap,i)
  {
    #pragma omp for schedule(dynamic)
  #endif
    for(i=0;i<j;++i)
    {
      #pragma omp critical
      {
      std::cout << "i=" << i+1 << ", " 
                << "period=" << gaps[i].begin()->enclosure->dimension() 
                << ", length=" << gaps[i].size() 
                << ", c=" << Interval(gaps[i].begin()->c.rightBound(),gaps[i].rbegin()->c.rightBound()) << std::endl;
      }
      try{      
        subdivideContainer(logMap[omp_get_thread_num()],gaps[i],omp_get_thread_num());
      }catch(std::exception& e)
      {
        std::cout << "exception: " << e.what() << std::endl;
        exit(0);
      }
      catch(const char * s)
      {
        std::cout << "exception: " << s << std::endl;
        exit(0);
      }
    }
  #ifdef __USE_PARALLEL__
  }
  #endif

  for(i=0;i<j;++i)
    c.insert(gaps[i].begin(),gaps[i].end());
}

